# coding=utf-8

# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test utils for Ironic Managers."""

import threading
from unittest import mock

import futurist
from futurist import periodics
from oslo_config import cfg
from oslo_utils import strutils
from oslo_utils import uuidutils

from ironic.common import driver_factory
from ironic.common import exception
from ironic.common import hash_ring
from ironic.common import states
from ironic.conductor import manager
from ironic import objects

CONF = cfg.CONF


class CommonMixIn(object):
    @staticmethod
    def _create_node(**kwargs):
        attrs = {'id': 1,
                 'uuid': uuidutils.generate_uuid(),
                 'power_state': states.POWER_OFF,
                 'target_power_state': None,
                 'maintenance': False,
                 'reservation': None}
        attrs.update(kwargs)
        node = mock.Mock(spec_set=objects.Node)
        for attr in attrs:
            setattr(node, attr, attrs[attr])
        return node

    def _create_task(self, node=None, node_attrs=None):
        if node_attrs is None:
            node_attrs = {}
        if node is None:
            node = self._create_node(**node_attrs)
        task = mock.Mock(spec_set=['node', 'release_resources',
                                   'spawn_after', 'process_event',
                                   'driver', 'context'])
        task.node = node
        return task

    def _get_nodeinfo_list_response(self, nodes=None):
        if nodes is None:
            nodes = [self.node]
        elif not isinstance(nodes, (list, tuple)):
            nodes = [nodes]
        return [tuple(getattr(n, c) for c in self.columns) for n in nodes]

    def _get_acquire_side_effect(self, task_infos):
        """Helper method to generate a task_manager.acquire() side effect.

        This accepts a list of information about task mocks to return.
        task_infos can be a single entity or a list.

        Each task_info can be a single entity, the task to return, or it
        can be a tuple of (task, exception_to_raise_on_exit). 'task' can
        be an exception to raise on __enter__.

        Examples: _get_acquire_side_effect(self, task): Yield task
                  _get_acquire_side_effect(self, [task, enter_exception(),
                                                  (task2, exit_exception())])
                       Yield task on first call to acquire()
                       raise enter_exception() in __enter__ on 2nd call to
                           acquire()
                       Yield task2 on 3rd call to acquire(), but raise
                           exit_exception() on __exit__()
        """
        tasks = []
        exit_exceptions = []
        if not isinstance(task_infos, list):
            task_infos = [task_infos]
        for task_info in task_infos:
            if isinstance(task_info, tuple):
                task, exc = task_info
            else:
                task = task_info
                exc = None
            tasks.append(task)
            exit_exceptions.append(exc)

        class FakeAcquire(object):
            def __init__(fa_self, context, node_id, *args, **kwargs):
                # We actually verify these arguments via
                # acquire_mock.call_args_list(). However, this stores the
                # node_id so we can assert we're returning the correct node
                # in __enter__().
                fa_self.node_id = node_id

            def __enter__(fa_self):
                task = tasks.pop(0)
                if isinstance(task, Exception):
                    raise task
                # NOTE(comstud): Not ideal to throw this into
                # a helper, however it's the cleanest way
                # to verify we're dealing with the correct task/node.
                if strutils.is_int_like(fa_self.node_id):
                    self.assertEqual(fa_self.node_id, task.node.id)
                else:
                    self.assertEqual(fa_self.node_id, task.node.uuid)
                return task

            def __exit__(fa_self, exc_typ, exc_val, exc_tb):
                exc = exit_exceptions.pop(0)
                if exc_typ is None and exc is not None:
                    raise exc

        return FakeAcquire


class ServiceSetUpMixin(object):
    def setUp(self):
        super(ServiceSetUpMixin, self).setUp()
        self.hostname = 'test-host'
        # Relies upon the default number of "NodeLocked" retries as
        # in unit testing, sqlite is not operated in a transactional
        # way and utilizes asynchronous IO. Locking, in particular, can
        # detect this, and it can cause some false or delayed inpressions
        # of lock status, causing lock failures.
        self.config(node_locked_retry_attempts=3, group='conductor')
        self.config(node_locked_retry_interval=0, group='conductor')

        self.service = manager.ConductorManager(self.hostname, 'test-topic')

    def _stop_service(self):
        try:
            objects.Conductor.get_by_hostname(self.context, self.hostname)
        except exception.ConductorNotFound:
            return
        self.service.del_host()

    def _start_service(self, start_periodic_tasks=False, start_consoles=True,
                       start_allocations=True):
        """Stand up a service, much like conductor base_manager.

        Ironic is a complex service, and the reality is that threading
        makes things far more complicated.

        The fun thing is that it is not actually that more complicated,
        but that we need to do things sanely and different for service
        startup than we need to do to predicate test setup. Largely around
        database initialization and thread usage in testing, otherwise we
        create unsuitable conditions for the tests in the test runners.

        Translation: If your adding any functionality here for tests,
        ensure a similar piece goes into ironic/conductor/base_manager.py.
        """

        self.service._shutdown = threading.Event()

        # Test class structure sets up self.dbapi, attaching it to
        # self.service for executing code to be able to leverage
        self.service.dbapi = self.dbapi
        if (CONF.rpc_transport == 'json-rpc'
                and CONF.json_rpc.port != 8089):
            self.service.host = f'{CONF.host}:{CONF.json_rpc.port}'
            self.hostname = CONF.host

        hardware_types = driver_factory.hardware_types()
        driver_factory.NetworkInterfaceFactory()
        driver_factory.StorageInterfaceFactory()
        hardware_type_names = list(hardware_types)

        # TODO(TheJulia): This should be largely handled for the specific
        # tests which need the fallack behavior.
        try:
            self.service.conductor = objects.Conductor.register(
                None, self.service.host, hardware_type_names,
                CONF.conductor.conductor_group)
        except exception.ConductorAlreadyRegistered:
            # This conductor was already registered and did not shut down
            # properly, so log a warning and update the record.
            self.conductor = objects.Conductor.register(
                None, self.service.host, hardware_type_names,
                CONF.conductor.conductor_group, update_existing=True)

        self.service._register_and_validate_hardware_interfaces(hardware_types)

        # Explicitly create some executors to handle threads from tasks.
        self.service._executor = futurist.SynchronousExecutor()
        self.service._reserved_executor = futurist.SynchronousExecutor()

        # Create a hash ring
        self.service.ring_manager = hash_ring.HashRingManager()

        # The next two steps are items expected by tests as related to
        # service startup in BaseConductorManager's prepare_host method
        self.dbapi.clear_node_target_power_state(self.service.host)
        self.dbapi.clear_node_reservations_for_conductor(self.service.host)

        if not start_periodic_tasks:
            with mock.patch.object(periodics, 'PeriodicWorker', autospec=True):
                self.service._collect_periodic_tasks(None)
                if start_allocations:
                    self.service._spawn_worker(
                        self.service._resume_allocations, None)
        else:
            self.service._collect_periodic_tasks(None)

        # Misc expectations.
        self.service._keepalive_evt = threading.Event()

        # Ideally, we should move this to the tests which need it.
        if CONF.conductor.enable_mdns:
            self.service._publish_endpoint()


def mock_record_keepalive(func_or_class):
    return mock.patch.object(
        manager.ConductorManager,
        '_conductor_service_record_keepalive',
        lambda _: None)(func_or_class)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
Middleware to log API request details including timing, status codes,
and request information for debugging purposes.
"""
import collections.abc
import time

from oslo_log import log

LOG = log.getLogger('ironic.api')


def get_real_ip(environ):
    """Safely retrieves the real IP address from a WSGI request."""
    # Check for the X-Forwarded-For header, which can contain a list of IPs.
    # The client's IP is the first one in the list.
    if 'HTTP_X_FORWARDED_FOR' in environ:
        # The header value is a comma-separated string, e.g., "client, proxy"
        # We take the first IP, which is the original client.
        return environ['HTTP_X_FORWARDED_FOR'].split(',')[0].strip()

    # If X-Forwarded-For is not present, check for X-Real-IP.
    elif 'HTTP_X_REAL_IP' in environ:
        return environ['HTTP_X_REAL_IP']

    # If no proxy headers are present, fall back to REMOTE_ADDR.
    else:
        return environ.get('REMOTE_ADDR')


class RequestLogMiddleware(object):
    """Middleware to log request details for debugging."""

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        # Capture request start time
        start_time = time.time()

        # Extract request details
        method = environ.get('REQUEST_METHOD', '')
        path = environ.get('PATH_INFO', '')
        query_string = environ.get('QUERY_STRING', '')

        # Build full path with query string if present
        full_path = path
        if query_string:
            full_path = f"{path}?{query_string}"

        # Capture response status
        status_code = None

        def logging_start_response(status, headers, exc_info=None):
            nonlocal status_code
            # Extract status code from status string (e.g., "200 OK")
            try:
                status_code = int(status.split(' ', 1)[0])
            except (ValueError, IndexError):
                status_code = 0
            return start_response(status, headers, exc_info)

        # Process the request
        try:
            response = self.app(environ, logging_start_response)
            # Ensure response is consumed if it's a generator
            if isinstance(response, collections.abc.Iterator):
                response = list(response)
            return response
        finally:
            # Calculate request duration
            duration = time.time() - start_time
            duration_ms = round(duration * 1000, 2)

            # Log the request details with request ID for traceability
            LOG.info("%(source_ip)s - %(method)s %(path)s - %(status)s (%("
                     "duration)sms)",
                     {'method': method,
                      'path': full_path,
                      'status': status_code or 'unknown',
                      'duration': duration_ms,
                      'source_ip': get_real_ip(environ) or 'unknown'})



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for ``ironic.networking.driver_factory``."""

from unittest import mock

from oslo_config import cfg

from ironic.common import exception
from ironic.networking.switch_drivers import driver_factory
from ironic import tests as tests_root


CONF = cfg.CONF


class BaseSwitchDriverFactoryTestCase(tests_root.base.TestCase):
    """Test behaviour common to switch driver factory classes."""

    def setUp(self):
        super(BaseSwitchDriverFactoryTestCase, self).setUp()
        self.factory = driver_factory.BaseSwitchDriverFactory()

    def test_set_enabled_drivers(self):
        self.config(group='ironic_networking', enabled_switch_drivers=['noop'])

        self.factory._set_enabled_drivers()

        self.assertEqual(['noop'], self.factory._enabled_driver_list)

    def test_set_enabled_drivers_with_duplicates(self):
        self.config(group='ironic_networking',
                    enabled_switch_drivers=['noop', 'noop'])

        exc = self.assertRaises(exception.ConfigInvalid,
                                self.factory._set_enabled_drivers)
        self.assertIn('noop', str(exc))
        self.assertIn('duplicated', str(exc))

    def test_set_enabled_drivers_with_empty_value(self):
        self.config(group='ironic_networking',
                    enabled_switch_drivers=['', 'noop'])

        exc = self.assertRaises(exception.ConfigInvalid,
                                self.factory._set_enabled_drivers)
        self.assertIn('empty', str(exc))
        self.assertIn('enabled_switch_drivers', str(exc))

    def test_init_extension_manager_no_enabled_drivers(self):
        self.factory._enabled_driver_list = []

        with mock.patch.object(
            driver_factory.LOG, 'info', autospec=True
        ) as mock_info:
            self.factory._init_extension_manager()

        mock_info.assert_called_once()

    @mock.patch('stevedore.NamedExtensionManager', autospec=True)
    def test_init_extension_manager_handles_runtime_error(
            self, mock_named_ext_mgr):
        # Reset extension manager state
        type(self.factory)._extension_manager = None
        type(self.factory)._enabled_driver_list = None
        self.config(group='ironic_networking', enabled_switch_drivers=['noop'])

        mock_named_ext_mgr.side_effect = RuntimeError(
            'No suitable drivers found')

        with mock.patch.object(
            driver_factory.LOG, 'warning', autospec=True
        ) as mock_warn:
            type(self.factory)._init_extension_manager()

        mock_warn.assert_called_once()
        self.assertIsNone(type(self.factory)._extension_manager)

    @mock.patch('stevedore.NamedExtensionManager', autospec=True)
    def test_init_extension_manager_unexpected_runtime_error(
            self, mock_named_ext_mgr):
        # Reset extension manager state
        type(self.factory)._extension_manager = None
        type(self.factory)._enabled_driver_list = None
        self.config(group='ironic_networking', enabled_switch_drivers=['noop'])

        mock_named_ext_mgr.side_effect = RuntimeError('boom')

        self.assertRaises(RuntimeError,
                          type(self.factory)._init_extension_manager)


class FactoryHelpersTestCase(tests_root.base.TestCase):
    """Tests for module-level helper functions."""

    def test_warn_if_unsupported(self):
        fake_extension = mock.Mock()
        fake_extension.obj.supported = False
        fake_extension.name = 'unsupported'

        with mock.patch.object(
            driver_factory.LOG, 'warning', autospec=True
        ) as mock_warn:
            driver_factory._warn_if_unsupported(fake_extension)

        mock_warn.assert_called_once()

    def test_warn_if_supported(self):
        fake_extension = mock.Mock()
        fake_extension.obj.supported = True

        with mock.patch.object(
            driver_factory.LOG, 'warning', autospec=True
        ) as mock_warn:
            driver_factory._warn_if_unsupported(fake_extension)

        mock_warn.assert_not_called()


class GlobalFactoryHelpersTestCase(tests_root.base.TestCase):
    """Tests for global helper functions providing factory access."""

    def setUp(self):
        super(GlobalFactoryHelpersTestCase, self).setUp()
        # Save and restore the global factory singleton
        original_factory = driver_factory._switch_driver_factory
        self.addCleanup(setattr, driver_factory, '_switch_driver_factory',
                        original_factory)
        driver_factory._switch_driver_factory = None

    def test_get_switch_driver_factory_singleton(self):
        factory1 = driver_factory.get_switch_driver_factory()
        factory2 = driver_factory.get_switch_driver_factory()
        self.assertIs(factory1, factory2)

    def test_get_switch_driver(self):
        factory = driver_factory.get_switch_driver_factory()

        with mock.patch.object(
            factory, 'get_driver', return_value='driver', autospec=True
        ) as mock_get:
            result = driver_factory.get_switch_driver('noop')

        mock_get.assert_called_once_with('noop')
        self.assertEqual('driver', result)

    def test_list_switch_drivers(self):
        factory = driver_factory.get_switch_driver_factory()

        with mock.patch.object(
            type(factory), 'names', new_callable=mock.PropertyMock,
            return_value=['noop']
        ):
            result = driver_factory.list_switch_drivers()

        self.assertEqual(['noop'], result)

    def test_switch_drivers(self):
        factory = driver_factory.get_switch_driver_factory()

        with mock.patch.object(
            factory, 'items', return_value=[('noop', 'driver')],
            autospec=True
        ):
            result = driver_factory.switch_drivers()

        self.assertEqual({'noop': 'driver'}, result)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add nodes.retired field

Revision ID: cd2c80feb331
Revises: ce6c4b3cf5a2
Create Date: 2020-01-16 12:51:13.866882

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'cd2c80feb331'
down_revision = 'ce6c4b3cf5a2'


def upgrade():
    op.add_column('nodes', sa.Column('retired', sa.Boolean(), nullable=True,
                                     server_default=sa.false()))
    op.add_column('nodes', sa.Column('retired_reason', sa.Text(),
                                     nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ironic.common.trait_based_networking.base as tbn
import ironic.common.trait_based_networking.grammar.parser as tbn_parser

from ironic.tests import base
import ironic.tests.unit.common.trait_based_networking.utils as tbn_test_utils

import lark

from dataclasses import dataclass


class TraitBasedNetworkingFilterParserTestCase(base.TestCase):
    def test_grammar_acceptable_to_lark(self):
        parser = lark.Lark(
            tbn_parser.FILTER_EXPRESSION_GRAMMAR,
            start=tbn_parser.FILTER_EXPRESSION_GRAMMAR_START_RULE)
        self.assertIsNotNone(parser)

    def test_filter_expression_parser(self):
        @dataclass
        class SubTestCase(object):
            description: str
            expression: str
            port_args: list[dict]
            # TODO(clif): Also test networks here
            expected_eval_result: list[bool]

        subtests = [
            SubTestCase(
                "Basic single expression",
                "port.vendor == 'vendor_string'",
                [{"vendor": "vendor_string"}],
                [True],
            ),
            SubTestCase(
                "Compound expressions with parenthesis",
                (
                    "port.vendor == 'green' "
                    "&& port.category == 'storage' "
                    "&& (port.physical_network =~ 'storage' "
                    "|| port.address == '192.168.1.1')"
                ),
                [
                    {
                        "vendor": "green",
                        "address": "192.168.1.1",
                        "physical_network": "storagenet",
                        "category": "storage",
                    },
                    {
                        "vendor": "green",
                        "address": "192.168.1.1",
                        "physical_network": "othernet",
                        "category": "storage",
                    },
                    {
                        "vendor": "brown",
                        "address": "192.168.1.1",
                        "physical_network": "storagenet",
                        "category": "storage_alpha",
                    },
                ],
                [True, True, False],
            ),
            SubTestCase(
                "Part 1 ensuring parens are respected",
                (
                    "(port.vendor == 'green' "
                    "&& port.category == 'test') "
                    "|| port.address == '192.168.1.1'"
                ),
                [
                    {
                        "vendor": "brown",
                        "address": "192.168.1.1",
                        "category": "prod",
                    },
                    {
                        "vendor": "brown",
                        "address": "192.168.1.1",
                        "category": "prod",
                    },
                ],
                [True, True],
            ),
            SubTestCase(
                "Part 2 ensuring parens are respected",
                (
                    "port.vendor == 'green' "
                    "&& (port.category == 'test' "
                    "|| port.address == '192.168.1.1')"
                ),
                [
                    {
                        "vendor": "brown",
                        "address": "192.168.1.1",
                        "category": "prod",
                    },
                    {
                        "vendor": "brown",
                        "address": "192.168.1.1",
                        "category": "prod",
                    },
                ],
                [False, False],
            ),
            # TODO(clif): Test case about network variables
        ]

        for subtest in subtests:
            with self.subTest(subtest=subtest):
                # Assert expression parses correctly.
                result = tbn.FilterExpression.parse(subtest.expression)
                self.assertIsNotNone(result)

                # Assert the transformation of the parse tree to
                # trait based networking objects renders back to the original
                # expression exactly.
                self.assertEqual(str(result), subtest.expression)

                self.assertEqual(
                    len(subtest.port_args), len(subtest.expected_eval_result)
                )

                # Assert the evaluation of the expression returns the correct
                # result.
                for i in range(0, len(subtest.port_args)):
                    self.assertEqual(
                        result.eval(
                            tbn.Port.from_ironic_port(
                                tbn_test_utils.FauxPortLikeObject(
                                    **subtest.port_args[i]
                                )
                            ),
                            tbn.Network("test_net_id", "test_network",
                                        frozenset()),
                        ),
                        subtest.expected_eval_result[i],
                    )



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from ironic.common.json_rpc import server as json_rpc_server
from ironic.common import rpc_service
from ironic.networking import manager as networking_manager
from ironic.tests import base as tests_base


CONF = cfg.CONF


class TestNetworkingRPCService(tests_base.TestCase):

    @mock.patch.object(json_rpc_server, "WSGIService", autospec=True)
    @mock.patch.object(
        rpc_service.objects_base, "IronicObjectSerializer", autospec=True
    )
    @mock.patch.object(rpc_service.context, "get_admin_context", autospec=True)
    def test_json_rpc_uses_networking_group(self, mock_ctx, mock_ser, mock_ws):
        CONF.set_override("rpc_transport", "json-rpc")
        # Ensure ironic networking group is registered and distinguishable
        CONF.set_override("port", 9999, group="ironic_networking_json_rpc")
        CONF.set_override("port", 8089, group="json_rpc")

        networking_manager.NetworkingManager(host="hostA")
        svc = rpc_service.BaseRPCService(
            "hostA", "ironic.networking.manager", "NetworkingManager"
        )

        # Trigger start path to build server
        with mock.patch.object(svc.manager, "prepare_host", autospec=True):
            with mock.patch.object(svc.manager, "init_host", autospec=True):
                svc._real_start()

        self.assertTrue(mock_ws.called)
        # Ensure conf_group was propagated to WSGIService
        _, kwargs = mock_ws.call_args
        self.assertEqual("ironic_networking_json_rpc",
                         kwargs.get("conf_group"))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
Redfish Inspect Interface
"""

from oslo_log import log
from oslo_utils import netutils
from oslo_utils import units
import sushy

from ironic.common import boot_modes
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common.inspection_rules import engine
from ironic.common import states
from ironic.common import utils
from ironic.conf import CONF
from ironic.drivers import base
from ironic.drivers.modules import inspect_utils
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic.drivers import utils as drivers_utils

LOG = log.getLogger(__name__)

PROCESSOR_INSTRUCTION_SET_MAP = {
    sushy.InstructionSet.ARM_A32: 'arm',
    sushy.InstructionSet.ARM_A64: 'aarch64',
    sushy.InstructionSet.IA_64: 'ia64',
    sushy.InstructionSet.MIPS32: 'mips',
    sushy.InstructionSet.MIPS64: 'mips64',
    sushy.InstructionSet.OEM: None,
    sushy.InstructionSet.X86: 'i686',
    sushy.InstructionSet.X86_64: 'x86_64'
}

BOOT_MODE_MAP = {
    sushy.BOOT_SOURCE_MODE_UEFI: boot_modes.UEFI,
    sushy.BOOT_SOURCE_MODE_BIOS: boot_modes.LEGACY_BIOS
}


class RedfishInspect(base.InspectInterface):

    def __init__(self):
        super().__init__()
        enabled_hooks = [x.strip()
                         for x in CONF.redfish.inspection_hooks.split(',')
                         if x.strip()]
        self.hooks = inspect_utils.validate_inspection_hooks("redfish",
                                                             enabled_hooks)

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return redfish_utils.COMMON_PROPERTIES.copy()

    def validate(self, task):
        """Validate the driver-specific Node deployment info.

        This method validates whether the 'driver_info' properties of
        the task's node contains the required information for this
        interface to function.

        This method is often executed synchronously in API requests, so it
        should not conduct long-running checks.

        :param task: A TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue on malformed parameter(s)
        :raises: MissingParameterValue on missing parameter(s)
        """
        redfish_utils.parse_driver_info(task.node)

    def inspect_hardware(self, task):
        """Inspect hardware to get the hardware properties.

        Inspects hardware to get the essential properties.
        It fails if any of the essential properties
        are not received from the node.

        :param task: a TaskManager instance.
        :raises: HardwareInspectionFailure if essential properties
                 could not be retrieved successfully.
        :returns: The resulting state of inspection.

        """
        system = redfish_utils.get_system(task.node)

        # get the essential properties and update the node properties
        # with it.
        inspected_properties = task.node.properties
        inventory = {}

        if system.memory_summary and system.memory_summary.size_gib:
            memory = system.memory_summary.size_gib * units.Ki
            inspected_properties['memory_mb'] = memory
            inventory['memory'] = {'physical_mb': memory}

        # match the inventory data of ironic-inspector / ironic-python-agent
        # to make existing inspection hooks and rules work by defaulting
        # the values
        inventory['cpu'] = {
            'count': 0,
            'architecture': '',
        }
        proc_info = self._get_processor_info(task, system)
        inventory['cpu'].update(proc_info)

        # TODO(etingof): should we respect root device hints here?
        local_gb = self._detect_local_gb(task, system)

        if local_gb:
            inspected_properties['local_gb'] = local_gb

        else:
            LOG.warning("Could not provide a valid storage size configured "
                        "for node %(node)s. Assuming this is a disk-less node",
                        {'node': task.node.uuid})
            inspected_properties['local_gb'] = 0

        if storages := system.storage or system.simple_storage:
            disks = list()
            for storage in storages.get_members():
                drives = storage.drives if hasattr(
                    storage, 'drives') else storage.devices
                for drive in drives:
                    disk = {}
                    disk['name'] = drive.name
                    disk['size'] = drive.capacity_bytes
                    disks.append(disk)

            inventory['disks'] = disks

        storage_controllers = self._get_storage_controllers(task, system)
        if storage_controllers:
            inventory['storage_controllers'] = storage_controllers

        inventory['interfaces'] = self._get_interface_info(task, system)

        pcie_devices = self._get_pcie_devices(system.pcie_devices)
        if pcie_devices:
            inventory['pci_devices'] = pcie_devices

        system_vendor = {}
        if system.model:
            system_vendor['product_name'] = str(system.model)

        if system.serial_number:
            system_vendor['serial_number'] = str(system.serial_number)

        if system.manufacturer:
            system_vendor['manufacturer'] = str(system.manufacturer)

        if system.uuid:
            system_vendor['system_uuid'] = str(system.uuid)

        if system_vendor:
            inventory['system_vendor'] = system_vendor

        if system.boot.mode:
            if not drivers_utils.get_node_capability(task.node, 'boot_mode'):
                capabilities = utils.get_updated_capabilities(
                    inspected_properties.get('capabilities', ''),
                    {'boot_mode': BOOT_MODE_MAP[system.boot.mode]})

                inspected_properties['capabilities'] = capabilities
            inventory['boot'] = {'current_boot_mode':
                                 BOOT_MODE_MAP[system.boot.mode]}

        self._create_ports(task, system)

        pxe_port_macs = self._get_pxe_port_macs(task)
        # existing data format only allows one mac so use that for now
        if pxe_port_macs:
            inventory['boot']['pxe_interface'] = pxe_port_macs[0]

        plugin_data = {}

        # Collect LLDP data from Redfish NetworkAdapter Ports
        # This method can be overridden by vendor-specific implementations
        lldp_raw_data = self._collect_lldp_data(task, system)
        if lldp_raw_data:
            plugin_data['parsed_lldp'] = lldp_raw_data
            LOG.info('Collected LLDP data for %(count)d interface(s) on '
                     'node %(node)s',
                     {'count': len(lldp_raw_data), 'node': task.node.uuid})

        inspect_utils.run_inspection_hooks(task, inventory, plugin_data,
                                           self.hooks, None)
        inspect_utils.store_inspection_data(task.node,
                                            inventory,
                                            plugin_data,
                                            task.context)
        engine.apply_rules(task, inventory, plugin_data, 'main')

        valid_keys = self.ESSENTIAL_PROPERTIES
        missing_keys = valid_keys - set(inspected_properties)
        if missing_keys:
            error = (_('Failed to discover the following properties: '
                       '%(missing_keys)s on node %(node)s') %
                     {'missing_keys': ', '.join(missing_keys),
                      'node': task.node.uuid})
            raise exception.HardwareInspectionFailure(error=error)

        task.node.properties = inspected_properties
        task.node.save()
        LOG.debug("Node properties for %(node)s are updated as "
                  "%(properties)s", {'properties': inspected_properties,
                                     'node': task.node.uuid})

        return states.MANAGEABLE

    def _create_ports(self, task, system):
        enabled_macs = redfish_utils.get_enabled_macs(task, system)
        if enabled_macs:
            inspect_utils.create_ports_if_not_exist(task, list(enabled_macs))
        else:
            LOG.warning("Not attempting to create any port as no NICs "
                        "were discovered in 'enabled' state for node "
                        "%(node)s: %(mac_data)s",
                        {'mac_data': enabled_macs, 'node': task.node.uuid})

    def _detect_local_gb(self, task, system):
        simple_storage_size = 0

        try:
            LOG.debug("Attempting to discover system simple storage size for "
                      "node %(node)s", {'node': task.node.uuid})
            if (system.simple_storage
                    and system.simple_storage.disks_sizes_bytes):
                simple_storage_size = [
                    size for size in system.simple_storage.disks_sizes_bytes
                    if size >= 4 * units.Gi
                ] or [0]

                simple_storage_size = simple_storage_size[0]

        except sushy.exceptions.SushyError as ex:
            LOG.debug("No simple storage information discovered "
                      "for node %(node)s: %(err)s", {'node': task.node.uuid,
                                                     'err': ex})

        storage_size = 0

        try:
            LOG.debug("Attempting to discover system storage volume size for "
                      "node %(node)s", {'node': task.node.uuid})
            if system.storage and system.storage.volumes_sizes_bytes:
                storage_size = [
                    size for size in system.storage.volumes_sizes_bytes
                    if size >= 4 * units.Gi
                ] or [0]

                storage_size = storage_size[0]

        except sushy.exceptions.SushyError as ex:
            LOG.debug("No storage volume information discovered "
                      "for node %(node)s: %(err)s", {'node': task.node.uuid,
                                                     'err': ex})

        try:
            if not storage_size:
                LOG.debug("Attempting to discover system storage drive size "
                          "for node %(node)s", {'node': task.node.uuid})
                if system.storage and system.storage.drives_sizes_bytes:
                    storage_size = [
                        size for size in system.storage.drives_sizes_bytes
                        if size >= 4 * units.Gi
                    ] or [0]

                    storage_size = storage_size[0]

        except sushy.exceptions.SushyError as ex:
            LOG.debug("No storage drive information discovered "
                      "for node %(node)s: %(err)s", {'node': task.node.uuid,
                                                     'err': ex})

        # NOTE(etingof): pick the smallest disk larger than 4G among available
        if simple_storage_size and storage_size:
            local_gb = min(simple_storage_size, storage_size)

        else:
            local_gb = max(simple_storage_size, storage_size)

        # Note(deray): Convert the received size to GiB and reduce the
        # value by 1 GB as consumers like Ironic requires the ``local_gb``
        # to be returned 1 less than actual size.
        return max(0, int(local_gb / units.Gi - 1))

    def _get_pxe_port_macs(self, task):
        """Get a list of PXE port MAC addresses.

        :param task: a TaskManager instance.
        :returns: Returns list of PXE port MAC addresses.
                  If cannot be determined, returns None.
        """
        return None

    def _get_interface_info(self, task, system):
        """Extract ethernet interface info."""

        ret = []
        if not system.ethernet_interfaces:
            return ret

        for eth in system.ethernet_interfaces.get_members():
            if not netutils.is_valid_mac(eth.mac_address):
                LOG.warning(_("Ignoring NIC address '%(address)s' for "
                              "interface %(inf)s on node %(node)s because it "
                              "is not a valid MAC"),
                            {'address': eth.mac_address,
                             'inf': eth.identity,
                             'node': task.node.uuid})
                continue
            mac_address = utils.normalize_mac(eth.mac_address)
            intf = {
                'mac_address': mac_address,
                'name': eth.identity
            }
            try:
                intf['speed_mbps'] = int(eth.speed_mbps)
            except Exception:
                pass
            ret.append(intf)
        return ret

    def _get_processor_info(self, task, system):
        # NOTE(JayF): Checking truthiness here is better than checking for None
        #             because if we have an empty list, we'll raise a
        #             ValueError.
        cpu = {}

        if not system.processors:
            return cpu

        if system.processors.summary:
            cpu['count'], _ = system.processors.summary

        processor = system.processors.get_members()[0]

        if processor.model is not None:
            cpu['model_name'] = str(processor.model)
        if processor.max_speed_mhz is not None:
            cpu['frequency'] = processor.max_speed_mhz
        cpu['architecture'] = PROCESSOR_INSTRUCTION_SET_MAP.get(
            processor.instruction_set) or ''

        return cpu

    def _get_storage_controllers(self, task, system):
        """Extract storage controller and drive information.

        Collects storage information from all Redfish Storage resources
        including their sub-controllers and drives.

        :param task: a TaskManager instance.
        :param system: Sushy system object.
        :returns: List of storage controller dictionaries.
        """
        controllers = []

        if not system.storage:
            return controllers

        try:
            members = system.storage.get_members()
        except sushy.exceptions.SushyError as ex:
            LOG.debug('Failed to get storage members for node %(node)s: '
                      '%(error)s', {'node': task.node.uuid, 'error': ex})
            return controllers

        for storage in members:
            controller = self._parse_storage_controller(task, storage)
            controllers.append(controller)

        return controllers

    @staticmethod
    def _enum_to_str(value):
        """Convert an enum value to string, or return as-is."""
        return value.value if hasattr(value, 'value') else value

    @staticmethod
    def _status_to_dict(status):
        """Convert a Sushy StatusField to a serializable dict."""
        if not status:
            return None
        result = {}
        if hasattr(status, 'health') and status.health is not None:
            result['health'] = RedfishInspect._enum_to_str(status.health)
        if hasattr(status, 'state') and status.state is not None:
            result['state'] = RedfishInspect._enum_to_str(status.state)
        if (hasattr(status, 'health_rollup')
                and status.health_rollup is not None):
            result['health_rollup'] = RedfishInspect._enum_to_str(
                status.health_rollup)
        return result or None

    def _parse_storage_controller(self, task, storage):
        """Parse a single storage resource.

        :param task: a TaskManager instance.
        :param storage: Sushy Storage object.
        :returns: Dictionary with storage controller info.
        """
        controller = {
            'id': str(storage.identity) if storage.identity else '',
            'name': storage.name if hasattr(storage, 'name') else None,
        }

        # Collect sub-controllers from deprecated StorageControllers list
        sub_controllers = []
        try:
            sc_list = storage.storage_controllers or []
        except (sushy.exceptions.SushyError, TypeError):
            sc_list = []
        for sc in sc_list:
            try:
                sc_info = {
                    'member_id': getattr(sc, 'member_id', None),
                    'name': getattr(sc, 'name', None),
                    'raid_types': [
                        rt.value for rt in
                        (getattr(sc, 'raid_types', None) or [])],
                    'speed_gbps': getattr(sc, 'speed_gbps', None),
                    'controller_protocols': [
                        p.value for p in
                        (getattr(sc, 'controller_protocols', None)
                         or [])],
                    'device_protocols': [
                        p.value for p in
                        (getattr(sc, 'device_protocols', None)
                         or [])],
                }
                status = getattr(sc, 'status', None)
                if status:
                    sc_info['status'] = self._status_to_dict(status)
                sub_controllers.append(sc_info)
            except (AttributeError, TypeError,
                    sushy.exceptions.SushyError):
                LOG.debug('Failed to parse sub-controller on node '
                          '%(node)s', {'node': task.node.uuid})
        controller['storage_controllers'] = sub_controllers

        # Collect drives
        drives = []
        if hasattr(storage, 'drives'):
            for drive in storage.drives:
                try:
                    drive_info = {
                        'name': drive.name,
                        'size': drive.capacity_bytes,
                        'id': getattr(drive, 'identity', None),
                        'media_type': getattr(drive, 'media_type',
                                              None),
                        'serial_number': getattr(drive, 'serial_number',
                                                 None),
                        'manufacturer': getattr(drive, 'manufacturer',
                                                None),
                        'model': getattr(drive, 'model', None),
                        'revision': getattr(drive, 'revision', None),
                    }
                    protocol = getattr(drive, 'protocol', None)
                    if protocol:
                        drive_info['protocol'] = self._enum_to_str(
                            protocol)
                    status = getattr(drive, 'status', None)
                    if status:
                        drive_info['status'] = self._status_to_dict(
                            status)
                    drives.append(drive_info)
                except (AttributeError, TypeError,
                        sushy.exceptions.SushyError):
                    LOG.debug('Failed to parse drive on node %(node)s',
                              {'node': task.node.uuid})
        controller['drives'] = drives

        return controller

    def _get_pcie_devices(self, pcie_devices_collection):
        """Extract PCIe device information from Redfish collection.

        :param pcie_devices_collection: Redfish PCIe devices collection
        :returns: List of PCIe device dictionaries
        """
        # Return empty list if collection is None
        if pcie_devices_collection is None:
            return []

        device_list = []

        # Process each PCIe device
        for pcie_device in pcie_devices_collection.get_members():
            # Skip devices that don't have functions
            if (not hasattr(pcie_device, 'pcie_functions')
                    or not pcie_device.pcie_functions):
                continue

            # Process each function on this device
            for pcie_function in pcie_device.pcie_functions.get_members():
                function_info = self._extract_function_info(pcie_function)
                if function_info:
                    device_list.append(function_info)

        return device_list

    def _extract_function_info(self, function):
        """Extract information from a PCIe function.

        :param function: PCIe function object
        :returns: Dictionary with function attributes
        """
        info = {}
        # Naming them same as in IPA for compatibility
        # IPA  has extra bus and numa_node_id which BMC doesn't have.
        if function.device_class is not None:
            info['class'] = str(function.device_class)
        if function.device_id is not None:
            info['product_id'] = function.device_id
        if function.vendor_id is not None:
            info['vendor_id'] = function.vendor_id
        if function.subsystem_id is not None:
            info['subsystem_id'] = function.subsystem_id
        if function.subsystem_vendor_id is not None:
            info['subsystem_vendor_id'] = function.subsystem_vendor_id
        if function.revision_id is not None:
            info['revision'] = function.revision_id
        return info

    def _collect_lldp_data(self, task, system):
        """Collect LLDP data from Redfish NetworkAdapter Ports.

        This method can be overridden by vendor-specific implementations
        to provide alternative LLDP data sources (e.g., Dell OEM endpoints).

        Default implementation uses standard Redfish LLDP data from
        Port.Ethernet.LLDPReceive via Sushy NetworkAdapter/Port resources.

        :param task: A TaskManager instance
        :param system: Sushy system object
        :returns: Dict mapping interface names to parsed LLDP data
                   Format: {'interface_name': {'switch_chassis_id': '..',
                                               'switch_port_id': '..'}}
        """
        parsed_lldp = {}

        try:
            # Check if chassis exists
            if not system.chassis:
                return parsed_lldp

            # Process each chassis
            for chassis in system.chassis:
                try:
                    # Get NetworkAdapters collection
                    network_adapters = (
                        chassis.network_adapters.get_members())
                except sushy.exceptions.SushyError as ex:
                    LOG.debug('Failed to get network adapters for chassis '
                              'on node %(node)s: %(error)s',
                              {'node': task.node.uuid, 'error': ex})
                    continue

                # Process each NetworkAdapter
                for adapter in network_adapters:
                    try:
                        # Get Ports collection using Sushy
                        ports = adapter.ports.get_members()
                    except sushy.exceptions.SushyError as ex:
                        LOG.debug('Failed to get ports for adapter '
                                  'on node %(node)s: %(error)s',
                                  {'node': task.node.uuid, 'error': ex})
                        continue

                    # Process each Port
                    for port in ports:
                        try:
                            # Check if LLDP data exists using Sushy
                            if (not port.ethernet
                                    or not port.ethernet.lldp_receive):
                                continue

                            lldp_receive = port.ethernet.lldp_receive

                            # Convert directly to parsed LLDP format
                            lldp_dict = self._convert_lldp_receive_to_dict(
                                lldp_receive)

                            if not lldp_dict:
                                continue

                            # Use port identity directly as interface name
                            if port.identity:
                                parsed_lldp[port.identity] = lldp_dict

                        except Exception as e:
                            LOG.debug('Failed to process LLDP data for port '
                                      '%(port)s on node %(node)s: %(error)s',
                                      {'port': port.identity,
                                       'node': task.node.uuid, 'error': e})
                            continue

        except Exception as e:
            LOG.warning('Failed to collect standard Redfish LLDP data for '
                        'node %(node)s: %(error)s',
                        {'node': task.node.uuid, 'error': e})
        return parsed_lldp

    def _convert_lldp_receive_to_dict(self, lldp_receive):
        """Convert Sushy LLDPReceive object directly to parsed dict format.

        :param lldp_receive: Sushy LLDPReceiveField object or dict
        :returns: Dict with parsed LLDP data or None
        """
        lldp_dict = {}

        # Chassis ID
        chassis_id = self._get_lldp_value(lldp_receive, 'chassis_id',
                                          'ChassisId')
        if chassis_id:
            lldp_dict['switch_chassis_id'] = chassis_id

        # Port ID
        port_id = self._get_lldp_value(lldp_receive, 'port_id', 'PortId')
        if port_id:
            lldp_dict['switch_port_id'] = port_id

        # System Name
        system_name = self._get_lldp_value(lldp_receive, 'system_name',
                                           'SystemName')
        if system_name:
            lldp_dict['switch_system_name'] = system_name

        # System Description
        system_description = self._get_lldp_value(lldp_receive,
                                                  'system_description',
                                                  'SystemDescription')
        if system_description:
            lldp_dict['switch_system_description'] = system_description

        # Management VLAN ID
        vlan_id = self._get_lldp_value(lldp_receive, 'management_vlan_id',
                                       'ManagementVlanId')
        if vlan_id:
            lldp_dict['switch_vlan_id'] = vlan_id

        return lldp_dict if lldp_dict else None

    def _get_lldp_value(self, lldp_receive, attr_name, json_key):
        """Get value from LLDP receive, handling both dict and object.

        :param lldp_receive: LLDP data (Sushy object or dict)
        :param attr_name: Sushy attribute name
        :param json_key: JSON property name (required)
        :returns: The value or None
        """
        # Being defensive to handle both Sushy object and dict
        if isinstance(lldp_receive, dict):
            return lldp_receive.get(json_key)
        else:
            return getattr(lldp_receive, attr_name, None)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Driver Configuration Adapter

This module provides functionality to translate user-friendly switch
configuration into driver-specific configuration formats. It allows users
to configure switches using a generic format while supporting multiple
switch driver implementations.
"""

import configparser
import glob
import os
import tempfile

from oslo_config import cfg
from oslo_log import log

from ironic.common import exception
from ironic.common.i18n import _

LOG = log.getLogger(__name__)

CONF = cfg.CONF


class NetworkingDriverAdapter:
    """Adapter for translating switch config to driver-specific format."""

    def __init__(self, driver_classes):
        """Initialize the driver adapter.

        :param driver_classes: Dictionary of driver name -> driver class.
            Classes must implement get_translator() classmethod.
        """
        self.driver_translators = {}
        self._register_translators(driver_classes)

    def _register_translators(self, driver_classes):
        """Register available driver translators.

        :param driver_classes: Dictionary of driver name -> driver class.
            Works with both driver classes and instances since
            get_translator() is a classmethod.
        """
        for name, driver_class in driver_classes.items():
            self.register_translator(name, driver_class.get_translator())

        LOG.debug(
            "Registered translators for drivers: %s",
            list(self.driver_translators.keys()),
        )

    def register_translator(self, driver_type, translator_instance):
        """Register a custom translator for a driver type.

        :param driver_type: String identifier for the driver type
        :param translator_instance: Instance of a translator class
        """
        self.driver_translators[driver_type] = translator_instance
        LOG.info(
            "Registered custom translator for driver type: %s", driver_type
        )

    def _validate_switch_config(self, switch_name, config):
        """Validate switch configuration has required fields.

        :param switch_name: Name of the switch
        :param config: Dictionary of configuration options
        :raises: NetworkError if validation fails
        """
        required_fields = [
            'driver_type',
            'device_type',
            'address',
            'username',
            'mac_address',
        ]
        missing_fields = [f for f in required_fields if f not in config]

        # Check for authentication: must have either password or key_file
        has_auth = 'password' in config or 'key_file' in config

        if missing_fields or not has_auth:
            error_parts = []
            if missing_fields:
                error_parts.append(
                    "missing required fields: %s" % ', '.join(missing_fields)
                )
            if not has_auth:
                error_parts.append(
                    "must specify either 'password' or 'key_file'"
                )

            raise exception.NetworkError(
                _("Invalid configuration for switch '%(switch)s': %(errors)s")
                % {'switch': switch_name, 'errors': '; '.join(error_parts)}
            )

    def preprocess_config(self, output_file):
        """Transform user config into driver-specific config files.

        Scans oslo.config for switch configurations and generates
        driver-specific config files that then get written to a driver-specific
        config file.

        :returns: Number of translations generated
        """
        try:
            if not os.path.exists(CONF.ironic_networking.switch_config_file):
                raise exception.NetworkError(
                    _("Switch configuration file %s does not exist")
                    % CONF.ironic_networking.switch_config_file
                )

            # Extract generic switch sections from config
            switch_sections = self._extract_switch_sections(
                CONF.ironic_networking.switch_config_file
            )

            if not switch_sections:
                LOG.debug(
                    "No user defined switch sections found in %s",
                    CONF.ironic_networking.switch_config_file,
                )
                return 0

            # Generate driver-specific configs
            translations = {}
            for switch_name, config in switch_sections.items():
                # Validate configuration before processing
                self._validate_switch_config(switch_name, config)

                driver_type = config["driver_type"]
                LOG.debug(
                    "Translating switch %s with driver type %s",
                    switch_name,
                    driver_type,
                )
                if driver_type in self.driver_translators:
                    translator = self.driver_translators[driver_type]
                else:
                    error_msg = (_("No driver translator registered for "
                                   "switch: %(switch_name)s, with driver "
                                   "type: %(driver_type)s") %
                                 {'switch_name': switch_name,
                                  'driver_type': driver_type})
                    raise exception.ConfigInvalid(error_msg=error_msg)

                translation = translator.translate_config(switch_name, config)
                if translation:
                    translations.update(translation)

            if translations:
                self._write_config_file(output_file, translations)
                CONF.reload_config_files()

            return len(translations)
        except Exception as e:
            LOG.exception("Failed to preprocess switch configuration: %s", e)
            raise exception.NetworkError(
                _("Configuration preprocessing failed: %s") % e
            )

    def _config_files(self):
        """Generate which yields all config files in the required order"""
        for config_file in CONF.config_file:
            yield config_file
        for config_dir in CONF.config_dir:
            config_dir_glob = os.path.join(config_dir, "*.conf")
            for config_file in sorted(glob.glob(config_dir_glob)):
                yield config_file

    def _extract_switch_sections(self, config_file):
        """Extract switch configuration sections from oslo.config.

        Looks for sections with names like:
        - [switch:switch_name]

        :returns: Dictionary of section_name -> config_dict
        """
        switch_sections = {}

        sections = {}
        parser = cfg.ConfigParser(config_file, sections)
        try:
            parser.parse()
        except Exception as e:
            LOG.warning("Failed to parse config file %s: %s", config_file, e)
            return {}

        for section_name, section_config in sections.items():
            if section_name.startswith("switch:"):
                switch_name = section_name[7:]
                # Get all key/value pairs in this section
                switch_sections[switch_name] = {
                    k: v[0] for k, v in section_config.items()
                }

        LOG.debug("Found %d switch sections", len(switch_sections))
        return switch_sections

    def _write_config_file(self, output_file, switch_configs):
        """Generate driver-specific configuration file.

        :param output_file: Path to the output file
        :param switch_configs: Dictionary of switch_name -> config_dict
        """
        # Create temp file in same directory as output file for atomic rename
        output_dir = os.path.dirname(output_file)
        temp_fd = None
        temp_path = None

        try:
            config = configparser.ConfigParser()

            # Add all sections and their key-value pairs
            for section_name, section_config in switch_configs.items():
                config.add_section(section_name)
                for key, value in section_config.items():
                    config.set(section_name, key, str(value))

            # Write to temporary file first
            temp_fd, temp_path = tempfile.mkstemp(
                dir=output_dir, prefix='.tmp_driver_config_', text=True
            )
            with os.fdopen(temp_fd, 'w') as f:
                temp_fd = None  # fdopen takes ownership
                f.write(
                    "# Auto-generated config for driver-specific switch "
                    "configurations\n"
                )
                f.write(
                    "# Generated from user defined switch configuration\n\n"
                )
                config.write(f)

            # Atomically move temp file to final location
            os.replace(temp_path, output_file)
            temp_path = None  # Successfully moved

            LOG.info("Generated driver config file: %s", output_file)

        except Exception:
            LOG.exception("Failed to generate config file")
            # Clean up temp file if it still exists
            if temp_fd is not None:
                try:
                    os.close(temp_fd)
                except OSError as cleanup_error:
                    LOG.debug("Failed to close temp file descriptor: %s",
                              cleanup_error)
            if temp_path is not None:
                try:
                    os.unlink(temp_path)
                except OSError as cleanup_error:
                    LOG.debug("Failed to remove temp file %s: %s",
                              temp_path, cleanup_error)
            raise

    def reload_configuration(self, output_file):
        """Reload and regenerate switch configuration files.

        This method re-extracts switch configurations from the config files
        and regenerates the driver-specific configuration files. It should
        be called when the switch configuration file has been modified.

        :param output_file: Path to the output file for driver-specific configs
        :returns: Number of translations generated
        :raises: NetworkError if configuration reload fails
        """
        LOG.info("Reloading switch configuration from config files")

        try:
            # Force oslo.config to reload configuration files
            CONF.reload_config_files()

            # Re-run the preprocessing steps
            count = self.preprocess_config(output_file)

            LOG.info(
                "Successfully reloaded switch configuration. "
                "Generated %d driver-specific config sections",
                count,
            )
            return count

        except Exception as e:
            LOG.error("Failed to reload switch configuration: %s", e)
            raise exception.NetworkError(
                _("Configuration reload failed: %s") % e
            )



#!/usr/bin/env python3
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import ironic.common.trait_based_networking.base as tbn
import ironic.common.trait_based_networking.config_file as tbn_config
import ironic.common.trait_based_networking.plan as tbn_plan

import ironic.tests.unit.common.trait_based_networking.utils as test_utils

import argparse
import yaml

arg_parser = argparse.ArgumentParser(
        prog="Trait Based Networking Simulator",
        description=("See how TBN will plan a network for a node given a set "
                     "of network traits, ports, portgroups, etc."))

arg_parser.add_argument('config_file')
arg_parser.add_argument('network_file')


def main():
    args = arg_parser.parse_args()

    cf = tbn_config.ConfigFile(args.config_file)
    print("Read Trait Based Networking configuration file "
          f"'{args.config_file}'")
    valid, reasons = cf.validate()
    if not valid:
        print(f"'{args.config_file}' is NOT valid!")
        for reason in reasons:
            print(f"ERROR: {reason}")
        print("Please correct these errors and try again.")
        return
    print("Configuration is valid.")

    cf.parse()
    print("Parsed configuration.")
    for trait in cf.traits():
        print(f"Got trait '{trait.name}' with {len(trait.actions)} action(s).")
    print("")

    print("Loading test set of network objects (ports, portgroups, networks) "
          f"from '{args.network_file}'")
    data = TestNetworkDataFile(args.network_file)
    print("Test set loaded.")
    print(f"Got {len(data.ports())} test ports.")
    print(f"Got {len(data.portgroups())} test portgroups.")
    print(f"Got {len(data.networks())} test networks.")
    print("")

    for trait in cf.traits():
        print(f"Planning based on trait '{trait.name}':")
        actions = tbn_plan.plan_network(trait,
                                        "fake_node_uuid",
                                        data.ports(),
                                        data.portgroups(),
                                        data.networks())
        print(f"Got {len(actions)} actions:")
        for action in actions:
            print(f"    {str(action)}")
        print("")


class TestNetworkDataFile(object):
    def __init__(self, filename):
        self._filename = filename
        self.read()

    def read(self):
        with open(self._filename, 'r') as file:
            self._contents = yaml.safe_load(file)

        for typename in ['ports', 'portgroups', 'networks']:
            assert typename in self._contents, (f"no {typename} were found in "
                                                 "the test data file!")

        self._ports = [
            tbn.Port.from_ironic_port(test_utils.FauxPortLikeObject(**port))
            for port in self._contents['ports']
        ]

        self._portgroups = [
            tbn.Portgroup.from_ironic_portgroup(
                test_utils.FauxPortLikeObject(**portgroup))
            for portgroup in self._contents['portgroups']
        ]

        self._networks = [
            tbn.Network(
                id=network['id'],
                name=network['name'],
                tags=frozenset(network.get('tags', ())))
            for network in self._contents['networks']
        ]

    def ports(self):
        return self._ports

    def portgroups(self):
        return self._portgroups

    def networks(self):
        return self._networks


if __name__ == "__main__":
    main()
